@@ -17,13 +17,20 @@ struct config
 	char name[256];                /* VFS/replication registriatio name */
 	unsigned long long failure_domain; /* User-provided failure domain */
 	unsigned long long int weight;     /* User-provided node weight */
+	char dir[1024];                    /* Data dir for on-disk database */
+	bool disk;                         /* Disk-mode or not */
+	int voters;                        /* Target number of voters */
+	int standbys;                      /* Target number of standbys */
 };
 
 /**
  * Initialize the config object with required values and set the rest to sane
  * defaults. A copy will be made of the given @address.
  */
-int config__init(struct config *c, dqlite_node_id id, const char *address);
+int config__init(struct config *c,
+		 dqlite_node_id id,
+		 const char *address,
+		 const char *dir);
 
 /**
  * Release any memory held by the config object.
