@@ -13,6 +13,7 @@
 #include "lib/serialize.h"
 
 #include "config.h"
+#include "id.h"
 #include "leader.h"
 #include "registry.h"
 #include "stmt.h"
@@ -30,19 +31,19 @@ struct gateway
 	struct raft *raft;           /* Raft instance */
 	struct leader *leader;       /* Leader connection to the database */
 	struct handle *req;          /* Asynchronous request being handled */
-	sqlite3_stmt *stmt;          /* Statement being processed */
-	bool stmt_finalize;          /* Whether to finalize the statement */
 	struct exec exec;            /* Low-level exec async request */
-	const char *sql;             /* SQL query for exec_sql requests */
 	struct stmt__registry stmts; /* Registry of prepared statements */
 	struct barrier barrier;      /* Barrier for query requests */
 	uint64_t protocol;           /* Protocol format version */
+	uint64_t client_id;
+	struct id_state random_state; /* For generating IDs */
 };
 
 void gateway__init(struct gateway *g,
 		   struct config *config,
 		   struct registry *registry,
-		   struct raft *raft);
+		   struct raft *raft,
+		   struct id_state seed);
 
 void gateway__close(struct gateway *g);
 
@@ -54,14 +55,57 @@ void gateway__leader_close(struct gateway *g, int reason);
 
 /**
  * Asynchronous request to handle a client command.
+ *
+ * We also use the handle as a place to save request-scoped data that we need
+ * to access from a callback.
  */
-typedef void (*handle_cb)(struct handle *req, int status, int type);
+typedef void (*handle_cb)(struct handle *req,
+			  int status,
+			  uint8_t type,
+			  uint8_t schema);
 struct handle
 {
-	void *data; /* User data */
-	int type;   /* Request type */
-	struct gateway *gateway;
+	/* User data. */
+	void *data;
+	/* Type code for this request. */
+	int type;
+	/* Schema version for this request. */
+	int schema;
+	/* Buffer where the response to this request will be written. */
 	struct buffer *buffer;
+	/* Cursor for reading the request. */
+	struct cursor cursor;
+	/* Database ID parsed from this request.
+	 *
+	 * This is used by handle_prepare. */
+	size_t db_id;
+	/* ID of the statement associated with this request.
+	 *
+	 * This is used by handle_prepare. */
+	size_t stmt_id;
+	/* SQL string associated with this request.
+	 *
+	 * This is used by handle_prepare, handle_query_sql, and handle_exec_sql
+	 * to save the provided SQL string across calls to leader__barrier and
+	 * leader__exec, since there's no prepared statement that can be saved
+	 * instead. In the case of handle_exec_sql, after preparing each
+	 * statement we update this field to point to the "tail" that has not
+	 * been prepared yet. */
+	const char *sql;
+	/* Prepared statement that will be queried to process this request.
+	 *
+	 * This is used by handle_query and handle_query_sql. */
+	sqlite3_stmt *stmt;
+	/* Number of times a statement parsed from this request has been
+	 * executed.
+	 *
+	 * This is used by handle_exec_sql, which parses zero or more statements
+	 * from the provided SQL string and executes them successively. Only if
+	 * at least one statement was executed should we fill the RESULT
+	 * response using sqlite3_last_insert_rowid and sqlite3_changes. */
+	unsigned exec_count;
+	/* Callback that will be invoked at the end of request processing to
+	 * write the response. */
 	handle_cb cb;
 };
 
@@ -72,14 +116,13 @@ struct handle
  * return an error if user code calls it and there's already a request in
  * progress.
  *
- * The @type parameter holds the request type code (e.g. #REQUEST_LEADER), the
- * @cursor parameter holds a cursor for reading the request payload, and the
- * @buffer parameter is a buffer for writing the response.
+ * The @type parameter holds the request type code (e.g. #REQUEST_LEADER), and
+ * the @buffer parameter is a buffer for writing the response.
  */
 int gateway__handle(struct gateway *g,
 		    struct handle *req,
 		    int type,
-		    struct cursor *cursor,
+		    int schema,
 		    struct buffer *buffer,
 		    handle_cb cb);
 
