@@ -1,5 +1,11 @@
 /* Implementation of the raft_uv_transport interface, proxied by a dqlite
- * connection. */
+ * connection.
+ *
+ * Instead of having raft instances connect to each other directly, we pass a
+ * custom connect function that causes dqlite to send a CONNECT request to the
+ * dqlite server where the destination raft instance is running. That server
+ * responds to the CONNECT request by forwarding the dqlite connection to its
+ * raft instance, after which the raft-to-raft connection is transparent. */
 #ifndef TRANSPORT_H_
 #define TRANSPORT_H_
 
@@ -7,6 +13,8 @@
 
 #include "../include/dqlite.h"
 
+int transportDefaultConnect(void *arg, const char *address, int *fd);
+
 int raftProxyInit(struct raft_uv_transport *transport, struct uv_loop_s *loop);
 
 void raftProxyClose(struct raft_uv_transport *transport);
