@@ -1,4 +1,4 @@
-#include "../../src/client.h"
+#include "../../src/client/protocol.h"
 #include "../../src/server.h"
 #include "../lib/client.h"
 #include "../lib/endpoint.h"
@@ -7,6 +7,7 @@
 #include "../lib/runner.h"
 #include "../lib/server.h"
 #include "../lib/sqlite.h"
+#include "../lib/util.h"
 
 /******************************************************************************
  *
@@ -17,7 +18,7 @@
 #define N_SERVERS 3
 #define FIXTURE                                \
 	struct test_server servers[N_SERVERS]; \
-	struct client *client;                 \
+	struct client_proto *client;           \
 	struct rows rows;
 
 #define SETUP                                                 \
@@ -31,7 +32,7 @@
 	test_server_network(f->servers, N_SERVERS);           \
 	for (i_ = 0; i_ < N_SERVERS; i_++) {                  \
 		struct test_server *server = &f->servers[i_]; \
-		test_server_start(server);                    \
+		test_server_start(server, params);            \
 	}                                                     \
 	SELECT(1)
 
@@ -52,26 +53,19 @@
 /* Use the client connected to the server with the given ID. */
 #define SELECT(ID) f->client = test_server_client(&f->servers[ID - 1])
 
-/* Wait a bounded time until server ID has applied at least the entry at INDEX */
-#define AWAIT_REPLICATION(ID, INDEX)							  \
-	do {										  \
-		struct timespec _start = {0};						  \
-		struct timespec _end = {0};						  \
-		clock_gettime(CLOCK_MONOTONIC, &_start);				  \
-		clock_gettime(CLOCK_MONOTONIC, &_end);					  \
-		while((f->servers[ID].dqlite->raft.last_applied < INDEX)		  \
-		      && ((_end.tv_sec - _start.tv_sec) < 2)  ) {			  \
-			clock_gettime(CLOCK_MONOTONIC, &_end);				  \
-		}									  \
-		munit_assert_ullong(f->servers[ID].dqlite->raft.last_applied, >=, INDEX); \
-	} while(0)
-
 /******************************************************************************
  *
  * join
  *
  ******************************************************************************/
 
+static char *bools[] = {"0", "1", NULL};
+
+static MunitParameterEnum membership_params[] = {
+    {"disk_mode", bools},
+    {NULL, NULL},
+};
+
 SUITE(membership)
 
 struct fixture
@@ -93,18 +87,18 @@ static void tearDown(void *data)
 	free(f);
 }
 
-TEST(membership, join, setUp, tearDown, 0, NULL)
+TEST(membership, join, setUp, tearDown, 0, membership_params)
 {
 	struct fixture *f = data;
 	unsigned id = 2;
 	const char *address = "@2";
-	unsigned stmt_id;
-	unsigned last_insert_id;
-	unsigned rows_affected;
+	uint32_t stmt_id;
+	uint64_t last_insert_id;
+	uint64_t rows_affected;
 
 	HANDSHAKE;
 	ADD(id, address);
-	ASSIGN(id, 1 /* voter */);
+	ASSIGN(id, DQLITE_VOTER);
 	OPEN;
 	PREPARE("CREATE TABLE test (n INT)", &stmt_id);
 	EXEC(stmt_id, &last_insert_id, &rows_affected);
@@ -112,31 +106,45 @@ TEST(membership, join, setUp, tearDown, 0, NULL)
 	EXEC(stmt_id, &last_insert_id, &rows_affected);
 
 	/* The table is visible from the new node */
+	TRANSFER(id, f->client);
 	SELECT(2);
 	HANDSHAKE;
 	OPEN;
 	PREPARE("SELECT * FROM test", &stmt_id);
 
 	/* TODO: fix the standalone test for remove */
-	SELECT(1);
-	REMOVE(id);
+	REMOVE(1);
 	return MUNIT_OK;
 }
 
-TEST(membership, transfer, setUp, tearDown, 0, NULL)
+struct id_last_applied
+{
+	struct fixture *f;
+	int id;
+	raft_index last_applied;
+};
+
+static bool last_applied_cond(struct id_last_applied arg)
+{
+	return arg.f->servers[arg.id].dqlite->raft.last_applied >=
+	       arg.last_applied;
+}
+
+TEST(membership, transfer, setUp, tearDown, 0, membership_params)
 {
 	struct fixture *f = data;
 	unsigned id = 2;
 	const char *address = "@2";
-	unsigned stmt_id;
-	unsigned last_insert_id;
-	unsigned rows_affected;
+	uint32_t stmt_id;
+	uint64_t last_insert_id;
+	uint64_t rows_affected;
 	raft_index last_applied;
-	struct client c_transfer; /* Client used for transfer requests */
+	struct client_proto c_transfer; /* Client used for transfer requests */
+	struct id_last_applied await_arg;
 
 	HANDSHAKE;
 	ADD(id, address);
-	ASSIGN(id, 1 /* voter */);
+	ASSIGN(id, DQLITE_VOTER);
 	OPEN;
 	PREPARE("CREATE TABLE test (n INT)", &stmt_id);
 	EXEC(stmt_id, &last_insert_id, &rows_affected);
@@ -157,26 +165,35 @@ TEST(membership, transfer, setUp, tearDown, 0, NULL)
 	PREPARE("INSERT INTO test(n) VALUES(1)", &stmt_id);
 	EXEC(stmt_id, &last_insert_id, &rows_affected);
 
-	AWAIT_REPLICATION(0, last_applied + 1);
+	await_arg.f = f;
+	await_arg.id = 0;
+	await_arg.last_applied = last_applied + 1;
+	AWAIT_TRUE(last_applied_cond, await_arg, 2);
 
 	return MUNIT_OK;
 }
 
 /* Transfer leadership away from a member that has a pending transaction */
-TEST(membership, transferPendingTransaction, setUp, tearDown, 0, NULL)
+TEST(membership,
+     transferPendingTransaction,
+     setUp,
+     tearDown,
+     0,
+     membership_params)
 {
 	struct fixture *f = data;
 	unsigned id = 2;
 	const char *address = "@2";
-	unsigned stmt_id;
-	unsigned last_insert_id;
-	unsigned rows_affected;
+	uint32_t stmt_id;
+	uint64_t last_insert_id;
+	uint64_t rows_affected;
 	raft_index last_applied;
-	struct client c_transfer; /* Client used for transfer requests */
+	struct client_proto c_transfer; /* Client used for transfer requests */
+	struct id_last_applied await_arg;
 
 	HANDSHAKE;
 	ADD(id, address);
-	ASSIGN(id, 1 /* voter */);
+	ASSIGN(id, DQLITE_VOTER);
 	OPEN;
 	PREPARE("CREATE TABLE test (n INT)", &stmt_id);
 	EXEC(stmt_id, &last_insert_id, &rows_affected);
@@ -204,26 +221,93 @@ TEST(membership, transferPendingTransaction, setUp, tearDown, 0, NULL)
 	PREPARE("INSERT INTO test(n) VALUES(2)", &stmt_id);
 	EXEC(stmt_id, &last_insert_id, &rows_affected);
 
-	AWAIT_REPLICATION(0, last_applied + 1);
+	await_arg.f = f;
+	await_arg.id = 0;
+	await_arg.last_applied = last_applied + 1;
+	AWAIT_TRUE(last_applied_cond, await_arg, 2);
 
 	return MUNIT_OK;
 }
 
-/* Transfer leadership back and forth from a member that has a pending transaction */
-TEST(membership, transferTwicePendingTransaction, setUp, tearDown, 0, NULL)
+struct fixture_id
+{
+	struct fixture *f;
+	int id;
+};
+
+static bool transfer_started_cond(struct fixture_id arg)
+{
+	return arg.f->servers[arg.id].dqlite->raft.transfer != NULL;
+}
+
+/* Transfer leadership away from a member and immediately try to EXEC a
+ * prepared SQL statement that needs a barrier */
+TEST(membership, transferAndSqlExecWithBarrier, setUp, tearDown, 0, NULL)
+{
+	int rv;
+	struct fixture *f = data;
+	unsigned id = 2;
+	const char *address = "@2";
+	uint32_t stmt_id;
+	uint64_t last_insert_id;
+	uint64_t rows_affected;
+	struct client_proto c_transfer; /* Client used for transfer requests */
+	struct fixture_id arg;
+
+	HANDSHAKE;
+	ADD(id, address);
+	ASSIGN(id, DQLITE_VOTER);
+	OPEN;
+	PREPARE("CREATE TABLE test (n INT)", &stmt_id);
+
+	/* Iniate transfer of leadership. This will cause a raft_barrier
+	 * failure while the node is technically still the leader, so the
+	 * gateway functionality that checks for leadership still succeeds. */
+	test_server_client_connect(&f->servers[0], &c_transfer);
+	HANDSHAKE_C(&c_transfer);
+	rv = clientSendTransfer(&c_transfer, 2, NULL);
+	munit_assert_int(rv, ==, 0);
+
+	/* Wait until transfer is started by raft so the barrier can fail. */
+	arg.f = f;
+	arg.id = 0;
+	AWAIT_TRUE(transfer_started_cond, arg, 2);
+
+	/* Force a barrier.
+	 * TODO this is hacky, but I can't seem to hit the codepath otherwise */
+	f->servers[0].dqlite->raft.last_applied = 0;
+
+	rv = clientSendExec(f->client, stmt_id, NULL, 0, NULL);
+	munit_assert_int(rv, ==, 0);
+	rv = clientRecvResult(f->client, &last_insert_id, &rows_affected, NULL);
+	munit_assert_int(rv, ==, DQLITE_CLIENT_PROTO_ERROR);
+
+	test_server_client_close(&f->servers[1], &c_transfer);
+	return MUNIT_OK;
+}
+
+/* Transfer leadership back and forth from a member that has a pending
+ * transaction */
+TEST(membership,
+     transferTwicePendingTransaction,
+     setUp,
+     tearDown,
+     0,
+     membership_params)
 {
 	struct fixture *f = data;
 	unsigned id = 2;
 	const char *address = "@2";
-	unsigned stmt_id;
-	unsigned last_insert_id;
-	unsigned rows_affected;
+	uint32_t stmt_id;
+	uint64_t last_insert_id;
+	uint64_t rows_affected;
 	raft_index last_applied;
-	struct client c_transfer; /* Client used for transfer requests */
+	struct client_proto c_transfer; /* Client used for transfer requests */
+	struct id_last_applied await_arg;
 
 	HANDSHAKE;
 	ADD(id, address);
-	ASSIGN(id, 1 /* voter */);
+	ASSIGN(id, DQLITE_VOTER);
 	OPEN;
 	PREPARE("CREATE TABLE test (n INT)", &stmt_id);
 	EXEC(stmt_id, &last_insert_id, &rows_affected);
@@ -251,7 +335,10 @@ TEST(membership, transferTwicePendingTransaction, setUp, tearDown, 0, NULL)
 	PREPARE("INSERT INTO test(n) VALUES(2)", &stmt_id);
 	EXEC(stmt_id, &last_insert_id, &rows_affected);
 
-	AWAIT_REPLICATION(0, last_applied + 1);
+	await_arg.f = f;
+	await_arg.id = 0;
+	await_arg.last_applied = last_applied + 1;
+	AWAIT_TRUE(last_applied_cond, await_arg, 2);
 
 	/* Transfer leadership back to original node, reconnect the client and
 	 * ensure queries can be executed. */
@@ -268,7 +355,8 @@ TEST(membership, transferTwicePendingTransaction, setUp, tearDown, 0, NULL)
 	PREPARE("INSERT INTO test(n) VALUES(3)", &stmt_id);
 	EXEC(stmt_id, &last_insert_id, &rows_affected);
 
-	AWAIT_REPLICATION(1, last_applied + 1);
+	await_arg.id = 1;
+	AWAIT_TRUE(last_applied_cond, await_arg, 2);
 
 	return MUNIT_OK;
 }
